/******************************************************************************
Description:

  Another example CIF_Comm program to test and demonstrate the C++
  implementation.  This example uses an event listener to detect when new
  connection requests and messages arrive.

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/examples/CIF_Comm_Event_Example.C,v $
  $Date: 1998/11/05 20:27:14 $
  $Revision: 1.5 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Toolkit.H"
#include "CIF_Comm.H"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/******************************************************************************
			     Function definitions
******************************************************************************/

class Example_Event_Listener
: public CIF_Toolkit_Event_Listener
{
    public:
    Example_Event_Listener()
    {
	globus_fifo_init(&l_conn_q);
	globus_fifo_init(&l_msg_q);
    }

    public:
    CIF_Comm_Connection *
    get_Connection()
    {
	return l_get_Connection(CIF_Toolkit_Timeout::INFINITY);
    }
    
    public:
    CIF_Comm_Connection *
    get_Connection(
	CIF_Toolkit_Timeout			timeout)
    {
	return l_get_Connection(timeout);
    }
    
    private:
    CIF_Comm_Connection *
    l_get_Connection(
	CIF_Toolkit_Timeout			timeout)
    {
	CIF_Toolkit_Monitor			monitor(l_mutex, l_conn_cond);

	if (globus_fifo_empty(&l_conn_q))
	{
	    while(1)
	    {
		monitor.wait(timeout);

		if (!globus_fifo_empty(&l_conn_q))
		{
		    break;
		}

		if (timeout == 0)
		{
		    return (CIF_Comm_Connection *) CIF_NULL;
		}
	    }
	}
	
	return (CIF_Comm_Connection *) globus_fifo_dequeue(&l_conn_q);
    }
    
    public:
    CIF_Comm_Message *
    get_Message()
    {
	return l_get_Message(CIF_Toolkit_Timeout::INFINITY);
    }
    
    public:
    CIF_Comm_Message *
    get_Message(
	CIF_Toolkit_Timeout			timeout)
    {
	return l_get_Message(timeout);
    }
    
    private:
    CIF_Comm_Message *
    l_get_Message(
	CIF_Toolkit_Timeout			timeout)
    {
	CIF_Toolkit_Monitor			monitor(l_mutex, l_msg_cond);

	if (globus_fifo_empty(&l_msg_q))
	{
	    while(1)
	    {
		monitor.wait(timeout);

		if (!globus_fifo_empty(&l_msg_q))
		{
		    break;
		}

		if (timeout == 0)
		{
		    return (CIF_Comm_Message *) CIF_NULL;
		}
	    }
	}
	
	return (CIF_Comm_Message *) globus_fifo_dequeue(&l_msg_q);
    }
    
    public:
    ~Example_Event_Listener()
    {
	while(!globus_fifo_empty(&l_conn_q))
	{
	    CIF_Comm_Connection *		conn;

	    conn = (CIF_Comm_Connection *) globus_fifo_dequeue(&l_conn_q);
	    conn->close();
	    delete conn;
	}
	
	while(!globus_fifo_empty(&l_msg_q))
	{
	    CIF_Comm_Message *			msg;

	    msg = (CIF_Comm_Message *) globus_fifo_dequeue(&l_conn_q);
	    delete msg;
	}

	globus_fifo_destroy(&l_conn_q);
	globus_fifo_destroy(&l_msg_q);
    }
    
    private:
    void
    handle_Event(
	CIF_Toolkit_Event &			event)
    {
	CIF_Toolkit_Monitor			monitor(l_mutex);
	
	if (event.is_Member(
	    CIF_Comm_Listener::Connection_Request_Event::CLASS_ID))
	{
	    CIF_Comm_Connection *		conn;
	    CIF_Comm_Connection::Properties	properties;
	    CIF_Comm_Listener *			lst;
	    CIF_Comm_Listener::
		Connection_Request_Event *	lst_event;

	    lst_event= (CIF_Comm_Listener::Connection_Request_Event *) &event;
	    
	    printf("handle_Event() got a Connection_Request_Event\n");

	    properties.set_Event_Listener(this);
	    
	    lst = lst_event->get_Listener();
	    conn = lst->get_Connection(properties, 0);
	    if (conn != CIF_NULL)
	    {
		printf("handle_Event() queued a new connection\n");
		globus_fifo_enqueue(&l_conn_q, (void *) conn);
		l_conn_cond.signal();
	    }
	    else
	    {
		printf("handle_Event() connection could not be formed "
		       "without blocking\n");
	    }
	}
	else if (event.is_Member(
	    CIF_Comm_Connection::Message_Received_Event::CLASS_ID))
	{
	    CIF_Comm_Connection *		conn;
	    CIF_Comm_Message *			msg;
	    CIF_Comm_Connection::
		Message_Received_Event *	conn_event;

	    conn_event= (CIF_Comm_Connection::Message_Received_Event *) &event;
	    
	    printf("handle_Event() got a Message_Received_Event\n");
	    msg = new CIF_Comm_Message();
	    conn = conn_event->get_Connection();
	    if (conn->receive(*msg, 0) == true)
	    {
		printf("handle_Event() queued a new message\n");
		globus_fifo_enqueue(&l_msg_q, (void *) msg);
		l_msg_cond.signal();
	    }
	    else
	    {
		printf("handle_Event() no message could be received "
		       "without blocking\n");
	    }
	}
	else
	{
	    printf("handle_Event() got a UNKNOWN EVENT!!!\n");
	}
    }

    CIF_Toolkit_Mutex				l_mutex;
    CIF_Toolkit_Cond				l_conn_cond;
    CIF_Toolkit_Cond				l_msg_cond;
    globus_fifo_t				l_conn_q;
    globus_fifo_t				l_msg_q;
};


/*
 * usage()
 */
void
usage(
    char *				exec_name)
{
    fprintf(stderr, "\nusage: %s [-c <nmsgs>] [-f] [-l] [-t <timeout>] "
	    "<-s|-r> <URL>\n\n", exec_name);
    fprintf(stderr, "    -b\tmessage size (bytes)\n");
    fprintf(stderr, "    -c\tmessage count\n");
    fprintf(stderr, "    -f\trepeat forever\n");
    fprintf(stderr, "    -l\tlistener\n");
    fprintf(stderr, "    -r\treciever\n");
    fprintf(stderr, "    -s\tsender\n");
    fprintf(stderr, "    -t\ttimeout value\n\n");

    exit(1);
}


/*
 * main()
 */
int    
main(
    int					argc,
    char *				argv[])
{
    const int				ARGC_REQ = 1;

    int					argn;
    int					msg_cnt = 1;
    int					msg_sz = 512;
    CIF_Bool				forever = false;
    CIF_Bool				listener = false;
    CIF_Bool				sender = false;
    CIF_Bool				receiver = false;
    int					timeout = -1;    
    char *				url;

    /*
     * Process arguments
     */
    argn = 0;
    while(++argn < argc - ARGC_REQ)
    {
	if (strcmp(argv[argn], "-b") == 0
	    && ++argn < argc - ARGC_REQ)
	{
	    msg_sz = atoi(argv[argn]);
	    continue;
	}
	if (strcmp(argv[argn], "-c") == 0
	    && ++argn < argc - ARGC_REQ)
	{
	    msg_cnt = atoi(argv[argn]);
	    continue;
	}
	else if (strcmp(argv[argn], "-f") == 0)
	{
	    forever = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-l") == 0)
	{
	    listener = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-s") == 0)
	{
	    sender = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-r") == 0)
	{
	    receiver = true;
	    continue;
	}
	else if (strcmp(argv[argn], "-t") == 0
		 && ++argn < argc - ARGC_REQ)
	{
	    timeout = atoi(argv[argn]);
	    continue;
	}
	else
	{
	    fprintf(stderr, "unknown argument: %s\n\n", argv[argn]);
	    usage(argv[0]);
	}
    }

    /*
     * Verify that all of the necessary arguments were specified
     */
    if (argn + ARGC_REQ == argc)
    {
	url = new char [strlen(argv[argn])+1];
	strcpy(url, argv[argn]);
    }
    else
    {
	usage(argv[0]);
    }

    if (!sender && !receiver)
    {
	usage(argv[0]);
    }
    

    /*
     * Activate CIF_Toolkit
     *
     * CIF_Toolkit::activate() and CIF_Comm::activate() must be executed before
     * any other CIF_Toolkit or CIF_Comm methods are called.  The activation
     * methods contain reference counting, so each module which uses one of the
     * CIF libraries can safely envoke the libraries activation method without
     * worrying about the potential problems associated with reactivation or
     * early deactivation.  However, it is important that all modules which
     * perform activation also perform deactivation when the have finished
     * using the CIF library.  Failing to perform deactivation can have
     * unpredicable results for some libraries.
     */
    CIF_Toolkit::activate();
    CIF_Comm::activate();

    /*
     * Establish a listener object and the wait for an incoming connection
     */
    Example_Event_Listener			event_listener;
    CIF_Comm_Listener *				cif_listener;
    CIF_Bool					rc;

    if (listener)
    {
	CIF_Comm_Listener::Properties		properties;

	properties.set_Event_Listener(&event_listener);
	
	printf("Establishing listener object for URL=%s\n", url);

	cif_listener = CIF_Comm_Factory::get_Listener(url, properties);

	printf("get_Listener() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(cif_listener != CIF_NULL);

	printf("\nListening at %s\n\n", cif_listener->get_URL());

	cif_listener->enable_Event_Listener();
    }
    

    do
    {
	CIF_Comm_Connection *			cif_conn;
	CIF_Comm_Connection::Properties		properties;
	    
	properties.set_Event_Listener(&event_listener);

	if (listener)
	{
	    /*
	     * Wait for a remote process contact the listener and form a
	     * new connection.
	     */
	    printf("Waiting for connection\n");

	    if (timeout >= 0)
	    {
		cif_conn = event_listener.get_Connection(timeout);
	    }
	    else
	    {
		cif_conn = event_listener.get_Connection();
	    }
	    
	    printf("get_Connection() %s\n",
		   CIF_Comm::status.get_Description());
	    assert(cif_conn != CIF_NULL);
	    assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	}
	else
	{
	    /*
	     * Attempt to establish a new connection
	     */
	    printf("Establishing connection object for URL=%s\n", url);
	    
	    if (timeout >= 0)
	    {
		cif_conn = CIF_Comm_Factory::get_Connection(url,
							    properties,
							    timeout);
	    }
	    else
	    {
		cif_conn = CIF_Comm_Factory::get_Connection(url, properties);
	    }

	    printf("get_Connection() %s\n",
		   CIF_Comm::status.get_Description());
	    assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	    assert(cif_conn != CIF_NULL);
	}
	printf("Connection established.\n");

	cif_conn->enable_Event_Listener();
	
	
	int					cnt;

	if (sender)
	{
	    CIF_Byte *				buf;
	    
	    buf = (CIF_Byte *) malloc(msg_sz);
	    
	    /*
	     * Send msg_cnt number of messages
	     */
	    printf("Sending %d messages\n", msg_cnt);
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		CIF_Comm_Message *		message;

		int				i;
		
		for (i = 0; i < msg_sz; i++)
		{
		    buf[i] = (cnt + i) % 256;
		}
		
		rc = cif_conn->send(buf, msg_sz, cnt);

		printf("send() %s\n",
		       CIF_Comm::status.get_Description());
		assert(rc == true);
		assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);

		while(1)
		{
		    message =
			event_listener.get_Message(CIF_Toolkit_Timeout(0));
		    if (message != CIF_NULL)
		    {
			printf("receive() got message; deleting\n");
			delete message;
		    }
		    else
		    {
			break;
		    }
		}
		
	    }

	    free(buf);
	}
	else
	{
	    /*
	     * Revceive msg_cnt number of messages
	     */
	    for (cnt = 0; cnt < msg_cnt; cnt++)
	    {
		int				msg_tag;
		int				msg_size;
		CIF_Byte *			msg_data;
		CIF_Bool			msg_valid;
		int				i;
		CIF_Comm_Message *		message;
	    
		message = event_listener.get_Message();
		msg_tag = message->get_Info().tag;
		msg_size = message->get_Info().size;
		msg_data = message->get_Data();

		printf(
		    "message received, size=%d, tag=%d\n", msg_size, msg_tag);

		if (msg_tag != cnt)
		{
		    printf("WARNING: expected message %d, got message %d\n",
			   cnt, msg_tag);
		}

		msg_valid = true;
		for (i = 0; i < msg_size; i++)
		{
		    if (msg_data[i] != ((cnt + i) % 256))
		    {
			msg_valid = false;
		    }
		}
		
		printf("get_Message() %s\n",
		       CIF_Comm::status.get_Description());
		assert(message != CIF_NULL);
		assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);

		delete message;
	    }
	}

	/*
	 * Close the connection and delete the connection object
	 */
	printf("Communication complete.  Closing connection.\n");
	
	rc = cif_conn->close();
	
	printf("close() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(rc == true);

	delete cif_conn;
    }
    while (forever);

    /*
     * Close the listener and delete the listener object
     */
    if (listener)
    {
	printf("Closing listener.\n");
	
	rc = cif_listener->close();
	
	printf("close() %s\n",
	       CIF_Comm::status.get_Description());
	assert(CIF_Comm::status == CIF_Comm_Status::SUCCESS);
	assert(rc == true);

	delete cif_listener;
    }
    
    /*
     * Detactivate CIF
     *
     * CIF_Comm::deactivate() must be called when the application has finished
     * using CIF_Comm.  Normally, this is done right before the application
     * exits.
     */
    CIF_Comm::deactivate();
    CIF_Toolkit::deactivate();

    exit(0);
}
