/******************************************************************************
CIF_I_Comm_Endpoint.H

Description:

  Classes to provide simple Nexus endpoint functionality

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/library/CIF_I_Comm_Endpoint.H,v $
  $Date: 1998/10/09 01:00:33 $
  $Revision: 1.6 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

#if !defined(CIF_INCLUDE_I_COMM_ENDPOINT)
#define CIF_INCLUDE_I_COMM_ENDPOINT 1

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Comm.H"

#include "globus_common.h"
#include "globus_nexus.h"


/******************************************************************************
			  Define CIF_I_Comm_Endpoint class
******************************************************************************/
class
CIF_I_Comm_Endpoint
: public CIF_Toolkit_Object
{
    public:
    class Properties
    : public CIF_I_Comm_Properties
    {
	public:
	Properties();

	public:
	void
	set_Protocol(
	    globus_nexus_proto_type_t		proto_type);

	public:
	void
	set_Protocol(
	    globus_nexus_proto_type_t		proto_type,
	    void *				proto_info_ptr,
	    int					proto_info_size);
	
	public:
	void
	get_Protocol(
	    globus_nexus_proto_type_t &		proto_type,
	    void * &				proto_info_ptr,
	    int &				proto_info_size) const;

	public:
	void
	enable_Attach();

	public:
	CIF_Bool
	is_Attach_Enabled() const;

	public:
	void
	set_Attach_Port(
	    const unsigned short		port);

	public:
	unsigned short
	get_Attach_Port() const;
	
	private:
	globus_nexus_proto_type_t		l_proto_type;
	void *					l_proto_info_ptr;
	int					l_proto_info_size;
	CIF_Bool				l_attach_enabled;
	unsigned short				l_attach_port;
    };

    
    public:
    class Message_Arrival_Event
    : public CIF_Toolkit_Event
    {
	public:
	Message_Arrival_Event(
	    CIF_I_Comm_Endpoint *		ep);
	    
	public:
	CIF_I_Comm_Endpoint *
	get_Endpoint() const;

	private:
	CIF_I_Comm_Endpoint *			l_ep;
    };
    

    /**************************************************************************
				    API methods
    **************************************************************************/

    /*
     * CIF_I_Comm_Endpoint::CIF_I_Comm_Endpoint()
     */
    public:
    CIF_I_Comm_Endpoint();

    /*
     * CIF_I_Comm_Endpoint::activate()
     */
    public:
    CIF_Bool
    activate(
	ARG_IN const Properties &		properties);

    /*
     * CIF_I_Comm_Endpoint::receive()
     */
    public:
    CIF_Bool
    receive(
	ARG_OUT globus_nexus_buffer_t &		nx_buf,
	ARG_IN const CIF_Toolkit_Timeout &	timeout);

    /*
     * CIF_I_Comm_Endpoint::poll()
     */
    public:
    CIF_Bool
    poll();

    /*
     * CIF_I_Comm_Endpoint::enable_Event_Listener()
     */
    public:
    CIF_Bool
    enable_Event_Listener();

    /*
     * CIF_I_Comm_Endpoint::disable_Event_Listener()
     */
    public:
    CIF_Bool
    disable_Event_Listener();

    /*
     * CIF_I_Comm_Endpoint::get_Attach_Hostname()
     */
    public:
    const char *
    get_Attach_Hostname() const;
    
    /*
     * CIF_I_Comm_Endpoint::get_Attach_Port()
     */
    public:
    unsigned short
    get_Attach_Port() const;
    
    /*
     * CIF_I_Comm_Endpoint::deactivate()
     */
    public:
    CIF_Bool
    deactivate();
    
    /*
     * CIF_I_Comm_Endpoint::~CIF_I_Comm_Endpoint()
     */
    public:
    ~CIF_I_Comm_Endpoint();
    

    /**************************************************************************
		       Internal methods and data structures
    **************************************************************************/

    /*
     * Objects which need access to our private methods
     */
    friend class CIF_I_Comm_Startpoint;


    private:
    static const int i_handle_Message_HID;


    /*
     * State of the endpoint object
     */
    private:
    struct Queue_Entry
    {
	globus_nexus_buffer_t			buf;
	CIF_Bool				event_dispatched;
	Queue_Entry *				next;
    };

    /*
     * State of the endpoint object
     */
    private:
    enum Object_State
    {
	OBJECT_STATE_ACTIVE,
	OBJECT_STATE_INACTIVE,
	OBJECT_STATE_DELETED
    };

    /*
     * State of the event_listener
     */
    private:
    enum Event_State
    {
	EVENT_STATE_DISABLED,
	EVENT_STATE_ENABLING,
	EVENT_STATE_ENABLED
    };

    /*
     * CIF_I_Comm_Endpoint::l_deactivate()
     */
    private:
    CIF_Bool
    l_deactivate(
	CIF_Toolkit_Monitor &			monitor);
    
    /*
     * CIF_I_Comm_Endpoint::bind_Startpoint()
     */
    private:
    CIF_Bool
    i_bind_Startpoint(
	ARG_INOUT
	globus_nexus_startpoint_t *		nx_sp);

    /*
     * CIF_I_Comm_Endpoint::l_verify_Object_State()
     */
    private:
    CIF_Comm_Status
    l_verify_Object_State(
	ARG_IN Object_State			desired_state);

    /*
     * CIF_I_Comm_Endpoint::handle_Message()
     */
    private: static
    int
    l_handle_Attach(
	ARG_IN void *				arg,
	ARG_IN char *				my_url,
	ARG_IN globus_nexus_startpoint_t *	nx_sp);

    /*
     * CIF_I_Comm_Endpoint::handle_Message()
     */
    private: static
    void
    l_handle_Message(
	globus_nexus_endpoint_t *		nx_ep,
	globus_nexus_buffer_t *			nx_buf,
	int					hid);

    /*
     * CIF_I_Comm_Endpoint::handle_Unknown()
     */
    private: static
    void
    l_handle_Unknown(
    globus_nexus_endpoint_t *			nx_ep,
    globus_nexus_buffer_t *			nx_buf,
    int						hid);

    /*
     * Internal data members
     */
    private:
    
    /* mutex to protect endpoint data structures */
    CIF_Toolkit_Mutex				l_mutex;

    /* endpoint for receiving connection establishment information */
    globus_nexus_endpoint_t			l_nx_ep;

    /* message queue */
    Queue_Entry * volatile			l_q_head;
    Queue_Entry * volatile			l_q_tail;
    CIF_Toolkit_Cond				l_q_cond;

    /* synchronization variables to ensure deactivation doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    volatile int				l_blocking_cnt;
    volatile CIF_Bool				l_deactivate_pending;
    CIF_Toolkit_Cond				l_deactivate_cond;
    
    /* port and enable flag for remote attach services */
    CIF_Bool					l_attach_enabled;
    unsigned short				l_attach_port;
    char *					l_attach_hostname;
    
    /*
     * object waiting for message events and a flag indicating that this object
     * is dispatching an event to the fore mentioned object
     */
    Event_State					l_event_state;
    CIF_Toolkit_Event_Listener *		l_event_listener;
    CIF_Toolkit_Cond				l_event_cond;
    CIF_Bool					l_dispatching_event;
    globus_thread_t				l_dispatching_thread;

    /* state of the endpoint object */
    volatile Object_State			l_object_state;

};

#endif /* CIF_INCLUDE_I_COMM_ENDPOINT */
